# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END

""" Module for Data-Faker poll mode plugin """

__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2024 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


import gc
import copy
import json
import logging

from foglamp.common import logger
from foglamp.plugins.common import utils
from foglamp.plugins.south.data_faker.utils import (
    weighted_random_categorical_value, weighted_random_continuous_value
)


_DEFAULT_CONFIG = {
    "plugin": {
        "description": "Generate fake data",
        "type": "string",
        "default": "data_faker",
        "readonly": "true"
    },
    "asset_datapoint_and_property": {
        "displayName": "Configuration",
        "description": "A list of asset, datapoints and property to according to which the data will be generated.",
        "type": "list",
        "items": "object",
        "default": '[ {"asset": "plc1", "datapoint" : "status", "feature_type" : "boolean"}, {"asset": "plc2", "datapoint" : "current", "feature_type" : "continuous", "minimum_value": "0.0", "maximum_value": "10.0"} ]',
        "properties": {
            "asset": {
                "displayName": "Asset name",
                "description": "The asset name to be generated.",
                "type": "string",
                "default": "asset1",
            },
            "datapoint": {
                "displayName": "Datapoint name",
                "description": "The datapoint name to be generated.",
                "type": "string",
                "default": "datapoint1",
            },
            "feature_type": {
                "displayName": "Feature Type",
                "description": "The kind of data that is to be generated for the feature.",
                "type": "enumeration",
                "options": [
                    "boolean",
                    "categorical",
                    "continuous",
                    "discrete",
                ],
                "default": "boolean",
            },
            "categories_value": {
                "validity": 'feature_type == "categorical"',
                "displayName": "Categories",
                "description": "The categories to generate. The categories names should be separated by comma.",
                "type": "string",
                "default": "low, mid, high",
            },
            "minimum_value": {
                "validity": 'feature_type == "continuous" ||  feature_type == "discrete"',
                "displayName": "Minimum",
                "description": "Minimum Value",
                "type": "float",
                "default": "0.0",
            },
            "maximum_value": {
                "validity": 'feature_type == "continuous" ||  feature_type == "discrete"',
                "displayName": "Maximum",
                "description": "Maximum Value",
                "type": "float",
                "default": "0.0",
            },
        }
    }
}

_LOGGER = logger.setup(__name__, level=logging.INFO)


def plugin_info():
    """Return information about the plugin.

    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    return {
        "name": "Data Faker Poll plugin",
        "version": "2.5.0",
        "mode": "poll",
        "type": "south",
        "interface": "1.0",
        "config": _DEFAULT_CONFIG
    }


def plugin_init(config):
    """Initialise the plugin.

    Args:
        config: JSON configuration document for the South plugin configuration category
    Returns:
        handle: JSON object to be used in future calls to the plugin
    Raises:
    """
    handle = copy.deepcopy(config)

    handle["dp_property"] = json.loads(handle["asset_datapoint_and_property"]["value"])
    for d in handle["dp_property"]:
        if d["feature_type"] == "categorical":
            d["categories_value"] = [e.strip() for e in d["categories_value"].split(",")]
        elif d["feature_type"] in ["continuous", "discrete"]:
            min_value, max_value = float(d["minimum_value"]), float(d["maximum_value"])
            if min_value > max_value:
                d["minimum_value"] = max_value
                d["maximum_value"] = max_value
            else:
                d["minimum_value"] = min_value
                d["maximum_value"] = max_value
            d["delta"] = (d["maximum_value"] - d["minimum_value"]) * 0.05  # 2%

    handle["last_value_map"] = None

    return handle


def plugin_poll(handle):
    """Extract data from the sensor and returns it in a JSON document as a Python dict.

    Available for poll mode only.
    Args:
        handle: handle returned by the plugin initialization call
    Returns:
        returns a sensor reading in a JSON document, as a Python dict, if it is available
        None - If no reading is available
    Raises:
        Exception
    """
    try:
        current_value = {}
        for d in handle["dp_property"]:

            last_value = (
                None
                if (
                    (handle["last_value_map"] is None) or
                    (d["asset"] not in handle["last_value_map"]) or
                    (d["datapoint"] not in handle["last_value_map"][d["asset"]])
                ) else
                handle["last_value_map"][d["asset"]][d["datapoint"]]
            )

            if d["asset"] not in current_value:
                current_value[d["asset"]] = {}

            if d["feature_type"] == "boolean":
                if last_value is None:
                    last_value = 0

                selected_value = weighted_random_categorical_value(
                    [0, 1],
                    last_value=last_value,
                    last_value_chance_multiplier = 10
                )

            elif d["feature_type"] == "categorical":
                if last_value is None:
                    last_value = d["categories_value"][0]

                selected_value = weighted_random_categorical_value(
                    d["categories_value"],
                    last_value=last_value,
                    last_value_chance_multiplier = 10
                )

            elif d["feature_type"] in ["continuous", "discrete"]:
                if last_value is None:
                    last_value = (d["minimum_value"] + d["maximum_value"]) / 2

                selected_value = weighted_random_continuous_value(
                    min_value=d["minimum_value"],
                    max_value=d["maximum_value"],
                    last_value=last_value,
                    delta_weight=d["delta"],
                    return_discrete_value=(
                        True if d["feature_type"] == "discrete" else False
                    )
                )

            current_value[d["asset"]][d["datapoint"]] = selected_value

        # creating the data for forwarding
        data, ts = [], utils.local_timestamp()
        for asset, reading in current_value.items():
            data.append({
                "readings": reading,
                "asset": asset,
                "id": 0,
                "ts": ts,
                "user_ts": ts
            })

        handle["last_value_map"] = current_value

        del current_value, asset, reading, selected_value, last_value, ts, d
        gc.collect()

    except (Exception, RuntimeError) as ex:
        _LOGGER.exception("Data-Faker Exception: {}".format(str(ex)))
        raise ex
    else:
        return data


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    Args:
        handle: handle returned by the plugin initialization call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _LOGGER.info("Old config for data-faker plugin {} \n new config {}".format(handle, new_config))
    new_handle = plugin_init(new_config)
    new_handle["last_value_map"] = handle["last_value_map"]
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup, to be called prior to the South plugin service being shut down.

    Args:
        handle: handle returned by the plugin initialization call
    Returns:
        plugin shutdown
    """
    _LOGGER.info("data-faker plugin shut down.")
